@@ -156,10 +156,7 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         for line in out.splitlines():
             if line:
                 info = line.split()
-                try:
-                    lines.append((info[0].split(',')[0], info[1]))
-                except IndexError:
-                    pass
+                lines.append((info[0].split(',')[0], info[1]))
         return lines
 
     def _get_iscsi_sessions(self):
@@ -167,45 +164,35 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         # entry: [tcp, [1], 192.168.121.250:3260,1 ...]
         return [entry[2] for entry in self._get_iscsi_sessions_full()]
 
-    def _get_ips_iqns_luns(self, connection_properties, discover=True,
-                           is_disconnect_call=False):
+    def _get_ips_iqns_luns(self, connection_properties, discover=True):
         """Build a list of ips, iqns, and luns.
 
-        Used when doing singlepath and multipath, and we have 4 cases:
+        Used only when doing multipath, we have 3 cases:
 
         - All information is in the connection properties
         - We have to do an iSCSI discovery to get the information
         - We don't want to do another discovery and we query the discoverydb
-        - Discovery failed because it was actually a single pathed attachment
 
         :param connection_properties: The dictionary that describes all
                                       of the target volume attributes.
         :type connection_properties: dict
         :param discover: Whether doing an iSCSI discovery is acceptable.
         :type discover: bool
-        :param is_disconnect_call: Whether this is a call coming from a user
-                                   disconnect_volume call or a call from some
-                                   other operation's cleanup.
-        :type is_disconnect_call: bool
         :returns: list of tuples of (ip, iqn, lun)
         """
-        # There are cases where we don't know if the local attach was done
-        # using multipathing or single pathing, so assume multipathing.
         try:
             if ('target_portals' in connection_properties and
                     'target_iqns' in connection_properties):
                 # Use targets specified by connection_properties
-                ips_iqns_luns = self._get_all_targets(connection_properties)
+                ips_iqns_luns = list(
+                    zip(connection_properties['target_portals'],
+                        connection_properties['target_iqns'],
+                        self._get_luns(connection_properties)))
             else:
                 method = (self._discover_iscsi_portals if discover
                           else self._get_discoverydb_portals)
                 ips_iqns_luns = method(connection_properties)
-        except exception.TargetPortalNotFound:
-            # Discovery failed, on disconnect this will happen if we
-            # are detaching a single pathed connection, so we use the
-            # connection properties to return the tuple.
-            if is_disconnect_call:
-                return self._get_all_targets(connection_properties)
+        except exception.TargetPortalsNotFound:
             raise
         except Exception:
             LOG.exception('Exception encountered during portal discovery')
@@ -321,6 +308,12 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
     def _get_transport(self):
         return self.transport
 
+    @staticmethod
+    def _get_luns(con_props, iqns=None):
+        luns = con_props.get('target_luns')
+        num_luns = len(con_props['target_iqns']) if iqns is None else len(iqns)
+        return luns or [con_props.get('target_lun')] * num_luns
+
     def _get_discoverydb_portals(self, connection_properties):
         """Retrieve iscsi portals information from the discoverydb.
 
@@ -468,7 +461,7 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
             root_helper=self._root_helper)
 
     @utils.trace
-    @synchronized('extend_volume', external=True)
+    @synchronized('extend_volume')
     def extend_volume(self, connection_properties):
         """Update the local kernel's size information.
 
@@ -491,10 +484,18 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
             raise exception.VolumePathsNotFound()
 
     @utils.trace
-    @synchronized('connect_volume', external=True)
+    @synchronized('connect_volume')
     def connect_volume(self, connection_properties):
         """Attach the volume to instance_name.
 
+        NOTE: Will retry up to three times to handle the case where c-vol
+        and n-cpu are both using os-brick to manage iSCSI sessions but they
+        are on the same node and using different locking directories. In this
+        case, even though this call is synchronized, they will be separate
+        locks and can still overlap with connect and disconnect. Since a
+        disconnect during an initial attach can't cause IO failure (the device
+        has not been made available yet), we just try the connection again.
+
         :param connection_properties: The valid dictionary that describes all
                                       of the target volume attributes.
         :type connection_properties: dict
@@ -622,23 +623,18 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         """
         device = hctl = None
         portal = props['target_portal']
-        try:
-            session, manual_scan = self._connect_to_iscsi_portal(props)
-        except Exception:
-            LOG.exception('Exception connecting to %s', portal)
-            session = None
+        session, manual_scan = self._connect_to_iscsi_portal(props)
+        do_scans = rescans > 0 or manual_scan
+        # Scan is sent on connect by iscsid, but we must do it manually on
+        # manual scan mode.  This scan cannot count towards total rescans.
+        if manual_scan:
+            num_rescans = -1
+            seconds_next_scan = 0
+        else:
+            num_rescans = 0
+            seconds_next_scan = 4
 
         if session:
-            do_scans = rescans > 0 or manual_scan
-            # Scan is sent on connect by iscsid, but we must do it manually on
-            # manual scan mode.  This scan cannot count towards total rescans.
-            if manual_scan:
-                num_rescans = -1
-                seconds_next_scan = 0
-            else:
-                num_rescans = 0
-                seconds_next_scan = 4
-
             data['num_logins'] += 1
             LOG.debug('Connected to %s', portal)
             while do_scans:
@@ -788,7 +784,7 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
                                         mpath)
 
     def _get_connection_devices(self, connection_properties,
-                                ips_iqns_luns=None, is_disconnect_call=False):
+                                ips_iqns_luns=None):
         """Get map of devices by sessions from our connection.
 
         For each of the TCP sessions that correspond to our connection
@@ -808,15 +804,14 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         connection properties (sendtargets was used on connect) so we may have
         to retrieve the info from the discoverydb.  Call _get_ips_iqns_luns to
         do the right things.
-
-        This method currently assumes that it's only called by the
-        _cleanup_conection method.
         """
         if not ips_iqns_luns:
-            # This is a cleanup, don't do discovery
-            ips_iqns_luns = self._get_ips_iqns_luns(
-                connection_properties, discover=False,
-                is_disconnect_call=is_disconnect_call)
+            if self.use_multipath:
+                # We are only called from disconnect, so don't discover
+                ips_iqns_luns = self._get_ips_iqns_luns(connection_properties,
+                                                        discover=False)
+            else:
+                ips_iqns_luns = self._get_all_targets(connection_properties)
         LOG.debug('Getting connected devices for (ips,iqns,luns)=%s',
                   ips_iqns_luns)
         nodes = self._get_iscsi_nodes()
@@ -854,7 +849,7 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         return device_map
 
     @utils.trace
-    @synchronized('connect_volume', external=True)
+    @synchronized('connect_volume')
     def disconnect_volume(self, connection_properties, device_info,
                           force=False, ignore_errors=False):
         """Detach the volume from instance_name.
@@ -876,12 +871,11 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         """
         return self._cleanup_connection(connection_properties, force=force,
                                         ignore_errors=ignore_errors,
-                                        device_info=device_info,
-                                        is_disconnect_call=True)
+                                        device_info=device_info)
 
     def _cleanup_connection(self, connection_properties, ips_iqns_luns=None,
                             force=False, ignore_errors=False,
-                            device_info=None, is_disconnect_call=False):
+                            device_info=None):
         """Cleans up connection flushing and removing devices and multipath.
 
         :param connection_properties: The dictionary that describes all
@@ -898,18 +892,13 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
                               ignore errors or raise an exception once finished
                               the operation.  Default is False.
         :param device_info: Attached device information.
-        :param is_disconnect_call: Whether this is a call coming from a user
-                                   disconnect_volume call or a call from some
-                                   other operation's cleanup.
-        :type is_disconnect_call: bool
         :type ignore_errors: bool
         """
         exc = exception.ExceptionChainer()
         try:
             devices_map = self._get_connection_devices(connection_properties,
-                                                       ips_iqns_luns,
-                                                       is_disconnect_call)
-        except exception.TargetPortalNotFound as exc:
+                                                       ips_iqns_luns)
+        except exception.TargetPortalsNotFound as exc:
             # When discovery sendtargets failed on connect there is no
             # information in the discoverydb, so there's nothing to clean.
             LOG.debug('Skipping cleanup %s', exc)
@@ -925,7 +914,8 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         was_multipath = (path_used.startswith('/dev/dm-') or
                          'mpath' in path_used)
         multipath_name = self._linuxscsi.remove_connection(
-            remove_devices, force, exc, path_used, was_multipath)
+            remove_devices, self.use_multipath, force, exc,
+            path_used, was_multipath)
 
         # Disconnect sessions and remove nodes that are left without devices
         disconnect = [conn for conn, (__, keep) in devices_map.items()
@@ -982,9 +972,9 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
             lines, _err = self._execute('cat', file_path, run_as_root=True,
                                         root_helper=self._root_helper)
 
-            for l in lines.split('\n'):
-                if l.startswith('InitiatorName='):
-                    return l[l.index('=') + 1:].strip()
+            for line in lines.split('\n'):
+                if line.startswith('InitiatorName='):
+                    return line[line.index('=') + 1:].strip()
         except putils.ProcessExecutionError:
             LOG.warning("Could not find the iSCSI Initiator File %s",
                         file_path)
@@ -1032,17 +1022,6 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         return ips, iqns
 
     def _connect_to_iscsi_portal(self, connection_properties):
-        """Safely connect to iSCSI portal-target and return the session id."""
-        portal = connection_properties['target_portal'].split(",")[0]
-        target_iqn = connection_properties['target_iqn']
-
-        lock_name = f'connect_to_iscsi_portal-{portal}-{target_iqn}'
-        method = synchronized(
-            lock_name, external=True)(self._connect_to_iscsi_portal_unsafe)
-        return method(connection_properties)
-
-    @utils.retry((exception.BrickException))
-    def _connect_to_iscsi_portal_unsafe(self, connection_properties):
         """Connect to an iSCSI portal-target an return the session id."""
         portal = connection_properties['target_portal'].split(",")[0]
         target_iqn = connection_properties['target_iqn']
@@ -1058,17 +1037,9 @@ class ISCSIConnector(base.BaseLinuxConnector, base_iscsi.BaseISCSIConnector):
         out, err = self._run_iscsiadm(connection_properties, (),
                                       check_exit_code=(0, 21, 255))
         if err:
-            out_new, err_new = self._run_iscsiadm(connection_properties,
-                                                  ('--interface',
-                                                   self._get_transport(),
-                                                   '--op', 'new'),
-                                                  check_exit_code=(0, 6))
-            if err_new:
-                # retry if iscsiadm returns 6 for "database failure"
-                LOG.debug("Retrying to connect to iSCSI portal %s", portal)
-                msg = (_("Encountered database failure for %s.") % (portal))
-                raise exception.BrickException(msg=msg)
-
+            self._run_iscsiadm(connection_properties,
+                               ('--interface', self._get_transport(),
+                                '--op', 'new'))
         # Try to set the scan mode to manual
         res = self._iscsiadm_update(connection_properties,
                                     'node.session.scan', 'manual',
